#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
A repository containing all of the constants frequently used in
this source separation stuff.
"""

DEFAULT_SAMPLE_RATE = 44100
DEFAULT_WIN_LEN_PARAM = 0.04
DEFAULT_BIT_DEPTH = 16
DEFAULT_MAX_VAL = 2 ** 16
EPSILON = 1e-16
MAX_FREQUENCY = DEFAULT_SAMPLE_RATE // 2
